/*-------------------------------------------------------------------*/
/* CBCROSS.C         Pasi Frnti                                     */
/*                                                                   */
/* Generic crossover for GA 										 */
/* Takes two solutions as input and returns a new one.				 */
/* Supports: Crossover, mutation and GLA iterations.                 */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName       "CBCROSS"
#define VersionNumber  "Version 0.01"
#define LastUpdated    "14.2.00"

/* ----------------------------------------------------------------- */

#include <assert.h>
#include <float.h>
#include <time.h>
#include <values.h>

#define  FACTFILE  "cbcross.fac"

#include "cross.h"

#include "parametr.c"
#include "cb.h"
#include "file.h"
#include "memctrl.h"
#include "random.h"
#include "solution.h"
#include "sortcb.h"

/*-------------------------- DOS special -----------------------------*/

#if defined(__MSDOS__)
#include <conio.h>
#endif


/* =================  T I M E   F U N C T I O N S  ================= */


static long SetWatch(long* start)
{
  time_t tmp;
  return( *start = (long)time(&tmp) );
}


/* ----------------------------------------------------------------- */


static void PrintTime(long start)
{
  time_t tmp;
  long t = (long)time(&tmp) - start;
  long h,m,s;

  s = t % 60;
  t = t / 60;
  m = t % 60;
  h = t / 60;

  switch( Value(QuietLevel) )
    {
    case 0:  break;
    case 1:
    case 2:
    case 3:
    case 4:  printf("  %li:%02li:%02li\n", h, m, s);
             break;
    default: printf("Time: %li:%02li:%02li\n", h, m, s);
             break;
    }
}


/*============================  P R I N T  ===============================*/


void PrintInfo(void)
{
  printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
  printf("CBCROSS - crosses two input solutions using Genetic Algorithm.\n");
  printf("Usage: %s [%coption] <training set> <solution 1> <solution 2>\n", ProgName, OPTION_SYMBOL);
  printf("For example: %s bridge solution1 solution2 newsolution \n", ProgName);
  printf("\n  Options:\n");
  PrintOptions();
  printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
}


/*-------------------------------------------------------------------*/


static void PrintOperatingInfo(char* TSName,
							   char* Name1,
							   char* Name2,
							   char* Name3)
{
  if( Value(QuietLevel) >= 2 )
    {
    printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
	printf("Training Set:        %s\n", TSName);
	printf("1st input solution:  %s\n", Name1);
	printf("2nd input solution:  %s\n", Name2);
	printf("New solution:        %s\n", Name3);
    printf("\n");
    PrintSelectedOptions();
    printf("\n");
    }
}


/*-------------------------------------------------------------------*/


static void PrintResult(long watch, double error)
{
  switch( Value(QuietLevel) )
	 {
     case 0:  break;
     case 1:
     case 2:  printf("%9.4f  ", error); break;
     default: printf("Distortion: %-9.4f\n", error); break;
	 }
  PrintTime(watch);
}


/*======================  SOLUTION HANDLING  ========================*/


static void OutputSolution(TRAININGSET*  TS,
						   char*		 CBName,
						   char*		 PAName,
						   SOLUTION*	 S)
{
  int ndups = DuplicatesInCodebook(&S->CB);

  if( ndups>0 && Value(QuietLevel)>=2 )
    {
    printf("WARNING: Final CB contains %i duplicates.\n", ndups);
    }
  SortCodebook(&S->CB, DATA_DESCENDING);
  WriteCodebook(CBName, &S->CB, Value(OverWrite));
  WritePartitioning(PAName, &S->P, TS, Value(OverWrite));
}


/*===========================  M A I N  ==============================*/


int main(int argc, char* argv[])
{
  TRAININGSET   TS;
  SOLUTION      Sold1;
  SOLUTION      Sold2;
  SOLUTION      Snew;
  char          TSName[64]  = { '\0' };
  char          Name1[64]   = { '\0' };
  char          Name2[64]   = { '\0' };
  char          Name3[64]   = { '\0' };
  char          CBName1[64] = { '\0' };
  char          CBName2[64] = { '\0' };
  char          CBName3[64] = { '\0' };
  char          PAName1[64] = { '\0' };
  char          PAName2[64] = { '\0' };
  char          PAName3[64] = { '\0' };
  double		Error;
  long          watch;
  
  ParameterInfo paraminfo[4] = { { TSName,	FormatNameTS, 0, INFILE  },
                                 { Name1,   "",           0, OUTFILE },
                                 { Name2,   "",           0, OUTFILE },
                                 { Name3,   "",           0, OUTFILE } };

  initrandom(0);
  ParseParameters(argc, argv, 4, paraminfo);
  PrintOperatingInfo(TSName, Name1, Name2, Name3);
  ReadTrainingSet(TSName, &TS);

  /* Copy file names but leave possible extensions out */
  PickFileName(Name1, CBName1);
  PickFileName(Name2, CBName2);
  PickFileName(Name3, CBName3);
  PickFileName(Name1, PAName1);
  PickFileName(Name2, PAName2);
  PickFileName(Name3, PAName3);

  /* Add the default extensions ".cb" and ".pa" to the file names  */
  CheckFileName(CBName1, FormatNameCB);
  CheckFileName(CBName2, FormatNameCB);
  CheckFileName(CBName3, FormatNameCB);
  CheckFileName(PAName1, FormatNamePA);
  CheckFileName(PAName2, FormatNamePA);
  CheckFileName(PAName3, FormatNamePA);
  
  /* Read input solutions from disk */
  ReadCodebook(CBName1, &Sold1.CB);
  ReadCodebook(CBName2, &Sold2.CB);
  ReadPartitioning(PAName1, &Sold1.P, &TS);
  ReadPartitioning(PAName2, &Sold2.P, &TS);
  Sold1.optimality = OPT_NONE;
  Sold2.optimality = OPT_NONE;
  CreateNewSolution(&TS, &Snew, 2*BookSize(&Sold1.CB));

  SetWatch(&watch);

  printf("\nCodebook 1:\n");
  PrintCodebook(&Sold1.CB);
  printf("\nCodebook 2:\n");
  PrintCodebook(&Sold1.CB);
  printf("\n");

  printf("Copy solution\n");
/* CopySolution(&Sold1, &Snew); */
  CrossSolutions(&TS, &Sold1, &Sold2, &Snew, Value(CrossingMethod));
  Mutate(&TS, &Snew, Value(Mutations));
  if( Value(GLAIterations) )
	 {
     IterateGLAForSolution(&TS, &Snew, Value(GLAIterations), MSE);
	 }
  Error = AverageErrorForSolution(&TS, &Snew.CB, &Snew.P, MSE);
  PrintResult(watch, PrintableError(Error,&(Snew.CB)));
  OutputSolution(&TS, CBName3, PAName3, &Snew);

  FreeCodebook(&TS);
  FreeSolution(&Sold1);
  FreeSolution(&Sold2);
  FreeSolution(&Snew);
  checkmemory();
  return( 0 );
}

