/*-------------------------------------------------------------------*/
/* CROSS.C         Pasi Frnti                                       */
/*                                                                   */
/* Crossover routines for Genetic Algorithm                          */
/* Supports the following:                                           */
/* CROSSOVER: Random, centroid distance, largest partitions,         */
/*            pairwise crossover, adaptive pairwise,                 */
/*			  PNN fast, adoption swap								 */
/* MUTATIONS: Random swap (single)									 */
/* FINE-TUNE: GLA													 */
/*-------------------------------------------------------------------*/

#define ProgName       "CROSS"
#define VersionNumber  "Version 0.01"
#define LastUpdated    "17.2.2000"

/* ----------------------------------------------------------------- */

#include <assert.h>
#include <float.h>
#include <time.h>
#include <values.h>

#include "pnnfast.h"

#include "cb.h"
#include "memctrl.h"
#include "random.h"
#include "solution.h"

/*--------------------------- Constants ------------------------------*/

#define  CBX(i)          ((i%2==0)? &S1->CB : &S2->CB)



/*======================  SOLUTION HANDLING  ========================*/


static int FreqToIndex(TRAININGSET* TS, int sum)
{
  int i=0,l=0;

  while (l<sum)  l += VectorFreq(TS, i++);
  return (i-1);
}


/*-------------------------------------------------------------------*/


static void GenerateRandomVector(TRAININGSET* TS, CODEBOOK* CB, int vec)
/* Frequencies are not taken into consideration. */
{
  int     i;
  YESNO   found=NO;
  int     index;

  while(!found)
     {
	 found=YES;
     index = irand(1,BookSize(TS)-1);
	 for (i=0; i<BookSize(CB); i++)
		if (EqualVectors(Vector(TS,index), Vector(CB,i), VectorSize(CB)))
           {
		   found=NO; break;
           }
     }
  CopyVector(Vector(TS,index), Vector(CB,vec), VectorSize(CB));
  VectorFreq(CB,vec)=1;
}


/*--------------------------------------------------------------------*/


static void SortByCentroidDistance(TRAININGSET* TS, CODEBOOK* CB)
{
  int		 i,j;
  llong*	 CBdist;
  llong 	 t;
  VECTORTYPE v;
  VECTORTYPE centroid;

  centroid = CreateEmptyVector(VectorSize(CB));
  CBdist   = (llong*) allocate(BookSize(CB) * sizeof(llong));

  /* calculate distances between code vectors and TS centroid */
  CodebookCentroid(TS, centroid);
  for (i=0; i<BookSize(CB); i++)
	  CBdist[i]=VectorDist(Vector(CB,i), centroid, VectorSize(TS));
    
  /* sort codebook according to the distances */
  for (i=0; i<BookSize(CB); i++)
	  for (j=0; j<i; j++)
		if (CBdist[j]>CBdist[i])
            {
			t=CBdist[i]; CBdist[i]=CBdist[j]; CBdist[j]=t;
			v=Vector(CB,i); Vector(CB,i)=Vector(CB,j); Vector(CB,j)=v;
            }

  deallocate(CBdist);
  FreeVector(centroid);
}


/*=================  CROSSOVER operations for CB  ====================*/


static int FindNearestNeighborVector(CODEBOOK* CB, int v)
{
  llong MinError=MAXLLONG;
  llong e;
  int	MinIndex=0;
  int	i;

  for (i=0; i<BookSize(CB); i++)
	 {
	 e = VectorDist(Vector(CB,i), Vector(CB,v), VectorSize(CB));
	 if( (e<MinError) && (i!=v) ) { MinError = e; MinIndex = i; }
	 }
  return( MinIndex );
}


/*-------------------------------------------------------------------*/

/* This is not needed
static int FindFirstNonEmpty(CODEBOOK* CB, PARTITIONING* P, int v)
{
  int	i;

  for (i=0; i<BookSize(CB); i++)
	 {
     if(UniqueVectors(P,i) && i!=v) return(i);
	 }
  PrintMessage("WARNING: No non-empty clusters found.\n");
  return( 0 );
}
*/

/*--------------------------------------------------------------------*/


static void CrossRandom(TRAININGSET* TS,
                        SOLUTION*    S1,
                        SOLUTION*    S2,
						SOLUTION*	 S3)
/* Generates Snew from S1 and S2 using RANDOM CROSSOVER method. */
{
  int      vsize=VectorSize(&S1->CB);
  int	   i,j,n;
  YESNO    found;
  double   error;

  for (i=0; i<BookSize(CBX(1)); i++)
    {
	found=NO;
	while(!found)
	   {
	   found=YES;
       n = irand(0,BookSize(CBX(1)));
	   for(j=0; j<i; j++)
          if(EqualVectors(Vector(CBX(i),n),Vector(&S3->CB,j),vsize))
             { found=NO; break; }
	   }
    CopyNode(&Node(CBX(i),n), &Node(&S3->CB,i), vsize);
	}
  error = GenerateOptimalPartitioning(TS, &S3->CB,&S3->P);
}


/*--------------------------------------------------------------------*/


static void ReplaceDuplicates(CODEBOOK* CB, CODEBOOK* CBreserve)
{
  int i,j;

  for ( i=0; i<BookSize(CB); i++ )
    for( j=i+1; j<BookSize(CB); j++ )
       {
       if( EqualVectors( Vector(CB,i), Vector(CB,j), VectorSize(CB) ))
          {
          GenerateRandomVector(CBreserve, CB, i);
          }
       }
}


/*--------------------------------------------------------------------*/


static void CrossCentroidDistance(TRAININGSET* TS,
                                  SOLUTION*    S1,
                                  SOLUTION*    S2,
								  SOLUTION*    S3)
/* Generates Snew from S1 and S2 using CENTROID DISTANCE method. */
{
  int   vsize = VectorSize(&S1->CB);
  int   bsize = BookSize(&S1->CB);
  int   i;
  int   r1 = irand(1,2);
  int   r2 = 3-r1;

  SortByCentroidDistance(TS, &S1->CB);
  SortByCentroidDistance(TS, &S2->CB);
  for (i=0; i<bsize/2; i++)
      CopyVector(Vector(CBX(r1),i), Vector(&S3->CB,i), vsize);
  for (i=bsize/2; i<bsize; i++)
      CopyVector(Vector(CBX(r2),i), Vector(&S3->CB,i), vsize);
  ReplaceDuplicates(&S3->CB,CBX(r1));
}


/*--------------------------------------------------------------------*/


static void CrossAdaptivePairwise(TRAININGSET* TS,
                                  SOLUTION*    S1,
                                  SOLUTION*    S2,
                                  SOLUTION*    S3)
/* Generates Snew from S1 and S2 using Adaptive Pairwise Swap */
{
  double e;
  llong  error;
  int    i, c;
  int    vsize = VectorSize(&S1->CB);
  int    bsize = BookSize(&S1->CB);

  CopyCodebook(&S1->CB, &S3->CB);
  CopyPartitioning(&S1->P, &S3->P);
  for(i=0; i<bsize; i++)
     if(irand(0,1)==0)
        {
        c = FindNearestVector( &Node(&S3->CB,i), &S2->CB, &error, i, EUCLIDEANSQ);
        CopyVector( Vector(&S2->CB,c), Vector(&S3->CB,i), vsize );
        }
  e = GenerateOptimalPartitioning(TS, &S3->CB, &S3->P);
  FillEmptyPartitions(TS, &S3->CB, &S3->P);
}


/*--------------------------------------------------------------------*/


static void CrossPNN(TRAININGSET* TS,
					 SOLUTION*	  S1,
					 SOLUTION*	  S2,
					 SOLUTION*	  S3)
/* Generates Snew from S1 and S2 using PNN fast */
{
  llong  d1, d2;
  int    i, p1, p2;
  int    booksize = BookSize(&S1->CB);
  int    vsize = VectorSize(&S1->CB);

  /* Combine codebooks: CB3 <- CB1 + CB2 */
  BookSize(&S3->CB) = 2*booksize;
  for(i=0; i<booksize; i++)
     CopyNode( &Node(&S1->CB,i), &Node(&S3->CB,i), vsize );
  for(i=0; i<booksize; i++)
     CopyNode( &Node(&S2->CB,i), &Node(&S3->CB,booksize+i), vsize );

  /* Combine partitions: P3 <- Best partitions from P1 and P2 */
  CopyPartitioning(&S1->P, &S3->P);
  PartitionCount(&S3->P) = 2*booksize;
  for(i=0; i<BookSize(TS); i++)
    {
    p1 = Map(&S1->P, i);
    p2 = Map(&S2->P, i);
    d1 = VectorDist(Vector(TS,i), Vector(&S1->CB,p1), vsize);
    d2 = VectorDist(Vector(TS,i), Vector(&S2->CB,p2), vsize);
    if(d2<d1) ChangePartition(TS, &S3->P, booksize+p2, i);
    }

  /* Calculate optimal codebook (2M) for combined partitioning */
  GenerateOptimalCodebook(TS, &S3->CB, &S3->P);
  S3->optimality = OPT_CB;

  /* Reduce codebook size from 2M -> M using PNN algorithm */
  PNNFast(TS, &S3->CB, &S3->P, booksize);
}


/*--------------------------------------------------------------------*/


static void CrossAdoptionSwap(TRAININGSET* TS,
							  SOLUTION*    S1,
							  SOLUTION*    S2,
                              SOLUTION*    S3)
/* Generates Snew from S1 and S2 using Adoption Swap */
{
  llong  error;
  int    i, c, guess;
  int    vsize = VectorSize(&S1->CB);
  int    bsize = BookSize(&S1->CB);

  CopyCodebook(&S1->CB, &S3->CB);
  CopyPartitioning(&S1->P, &S3->P);
  for(i=0; i<bsize; i++)
     {
     guess = Map(&S3->P,FirstVector(&S2->P,i));
     c = FindNearestVector( &Node(&S2->CB,i), &S3->CB, &error, guess, EUCLIDEANSQ);
     CopyVector( Vector(&S2->CB,i), Vector(&S3->CB,c), vsize );
     LocalRepartitioningGeneral(TS, &S3->CB, &S3->P, c, EUCLIDEANSQ);
     RepartitionDueToNewVectorGeneral(TS, &S3->CB, &S3->P, c, EUCLIDEANSQ);
     GenerateOptimalCodebook(TS, &S3->CB, &S3->P);
     S3->optimality = OPT_CB;
     }
}


/*--------------------------------------------------------------------*/


void CrossSolutions(TRAININGSET* TS,
					SOLUTION*	 S1,
					SOLUTION*	 S2,
                    SOLUTION*    S3,
                    int          Method)
{
  switch( Method )
    {
    case 0:  CopySolution(S1, S3); break;
	case 1:  CrossRandom(TS, S1, S2, S3); break;
	case 2:  CrossCentroidDistance(TS, S1, S2, S3); break;
    case 3:  /* LP - not yet implemented here !!! */
    case 4:  /* PC - not yet implemented here !!!  */
    case 5:  CrossAdaptivePairwise(TS, S1, S2, S3);  break;
	case 6:  /* PNN original - not supported !!! */
    case 7:  CrossPNN(TS, S1, S2, S3);  break;
    case 8:  CrossAdoptionSwap(TS, S1, S2, S3); break;
    default: break;
    }
}


/*===========================  MUTATION  =============================*/


void Mutate(TRAININGSET* TS, SOLUTION* S, int NumberOfMutations)
{
  int i;

  if( NumberOfMutations==0 ) return;
  for (i=0; i<=NumberOfMutations; i++)
	 {
     GenerateRandomVector( TS, &S->CB, irand(0,BookSize(&S->CB)-1) );
	 }
  S->optimality = OPT_NONE;
}


