/*-------------------------------------------------------------------*/
/* PNNFAST.C         Pasi Frnti                                     */
/*                                                                   */
/* Fast PNN method                                                   */
/*   Starts from initial solution                                    */
/*	 Reduces cluster size to FinalSize								 */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName	   "PNN-FAST"
#define VersionNumber  "Version 0.01"
#define LastUpdated    "14.2.2000"

/* ----------------------------------------------------------------- */

#include <assert.h>
#include <float.h>
#include <time.h>
#include <values.h>

#include "cb.h"
#include "memctrl.h"
#include "random.h"
#include "solution.h"

/*---------------------------  M i s c	------------------------------*/

#define  round(a)		((int) (a+0.5))
#define  swap(a,b)      { int t = a; a = b; b = t; }

/*----------------------- Type definitions --------------------------*/

typedef struct { llong  VectorDist;
                 int    Vector;
                 YESNO  Recalculate;
               } NEARESTTYPE;


/*====================================================================*/


static llong MergeDistortion(CODEBOOK* CB, int p1, int p2)
{
  llong   dist;
  double  factor;

  if( VectorFreq(CB,p1)==0 || VectorFreq(CB,p2)==0 )
     {
     dist = 0LL;
     }
  else
     {
     dist   = VectorDist(Vector(CB,p1), Vector(CB,p2), VectorSize(CB));
     factor = ((double)VectorFreq(CB,p1) * (double)VectorFreq(CB,p2)) /
              ((double)VectorFreq(CB,p1) + (double)VectorFreq(CB,p2));
     dist   = round( factor * (double) dist);
     }
  return(dist);
}


/*-------------------------------------------------------------------*/


static void FindNearestCluster(CODEBOOK* CB, int a, int* b, llong* dist)
{
  llong Dist;
  int i;

  (*dist)=MAXLLONG; (*b)=0;
  for( i=0; i<BookSize(CB); i++ )
     {
     if( a!=i )
        {
        Dist = MergeDistortion(CB, a, i);
        if( Dist < (*dist) )
           {
           (*dist) = Dist;
           (*b)    = i;
           }
        }
     }
}


/*-------------------------------------------------------------------*/


static void CalculateAllDistances(CODEBOOK* CB, NEARESTTYPE NN[])
{
  int i;

  for( i=0; i<BookSize(CB); i++ )
    {
    FindNearestCluster(CB, i, &(NN[i].Vector), &(NN[i].VectorDist));
    }
}


/*-------------------------------------------------------------------*/


static void FindMergedPair(CODEBOOK* CB, NEARESTTYPE NN[], int* a, int* b)
{
  llong MinDist=MAXLLONG;
  int i;

  for( i=0; i<BookSize(CB); i++ )
    {
    if( NN[i].VectorDist < MinDist )
       {
       (*a) = i;
       (*b) = NN[i].Vector;
       MinDist = NN[i].VectorDist;
       }
    }
  if( (*a)>(*b) ) swap((*a),(*b));
}


/*-------------------------------------------------------------------*/


static void CentroidOfTwoVectors(CODEBOOK* CB, BOOKNODE* cent, int p1, int p2)
{
  int k;
  int freqsum;

  freqsum = VectorFreq(CB,p1) + VectorFreq(CB,p2);

  /* Calculate mean values for centroid */
  for( k=0; k<VectorSize(CB); k++)
    {
	if( freqsum > 0 )
	cent->vector[k] = round( (double)(VectorFreq(CB, p1)*VectorScalar(CB, p1, k) +
                                      VectorFreq(CB, p2)*VectorScalar(CB, p2, k))
                             / (double)freqsum );
	else cent->vector[k] = 0;
    }
  cent->freq = freqsum;
}


/* ----------------------------------------------------------------- */


static void RemoveLastVector(CODEBOOK* CB, int a, int b, NEARESTTYPE NN[])
{
  int last = BookSize(CB)-1;
  int i;

  if( b != last )
    {
    CopyNode(&Node(CB,last), &Node(CB,b), VectorSize(CB));
    NN[b] = NN[last];
    for( i=0; i<BookSize(CB); i++ )
       {
       if( NN[i].Vector==last )  NN[i].Vector = b;
       }
    }
}


/*--------------------------------------------------------------------*/


static void RemoveEmptyPartitions(TRAININGSET*  TS,
                                  CODEBOOK*     CB,
                                  PARTITIONING* P,
                                  int           FinalSize)
{
  int i,last;

  for( i=0; i<BookSize(CB); i++ )
     {
     if( VectorFreq(CB,i)==0  &&  BookSize(CB)>FinalSize )
        {
        last = BookSize(CB)-1;
        CopyNode(&Node(CB,last), &Node(CB,i), VectorSize(CB));
        JoinPartitions(TS, P, i, last);
        BookSize(CB)--;
        i--;
        }
     }
}


/*--------------------------------------------------------------------*/


static void MarkClustersForRecalculation(CODEBOOK* CB,
                                         NEARESTTYPE NN[],
                                         int a, int b)
{
  int i;

  for( i=0; i<BookSize(CB); i++ )
    {
	NN[i].Recalculate = ( NN[i].Vector==a || NN[i].Vector==b ) ? YES : NO;
	}
}


/*--------------------------------------------------------------------*/


static void RecalculateDistances(CODEBOOK* CB, NEARESTTYPE NN[])
{
  int  i;
  int  nrec=0;

  for( i=0; i<BookSize(CB); i++ )
    {
    if( NN[i].Recalculate )
       {
       FindNearestCluster(CB, i, &(NN[i].Vector), &(NN[i].VectorDist));
       NN[i].Recalculate = NO;
       nrec++;
       }
    }
/* printf("Recalculated: %i / %i.\n", nrec, BookSize(CB)); */
}


/*-------------------------------------------------------------------*/


static void MergeVectors(TRAININGSET*  TS,
                         CODEBOOK*     CB,
                         PARTITIONING* P,
                         NEARESTTYPE   NN[],
                         int           a,
                         int           b)
{
  MarkClustersForRecalculation(CB, NN, a, b);
  CentroidOfTwoVectors(CB, &Node(CB, a), a, b);
  JoinPartitions(TS, P, a, b);
  RemoveLastVector(CB, a, b, NN);
  BookSize(CB)--;
}


/*-------------------------------------------------------------------*/


void PNNFast(TRAININGSET*	TS,
			 CODEBOOK*		CB,
			 PARTITIONING*	P,
			 int			FinalSize)
{
  NEARESTTYPE*  NN;
  int  a,b;

  RemoveEmptyPartitions(TS, CB, P, FinalSize);
  NN = (NEARESTTYPE*) allocate(BookSize(CB) * sizeof(NEARESTTYPE));
  CalculateAllDistances(CB, NN);
  while( BookSize(CB) > FinalSize )
     {
     FindMergedPair(CB, NN, &a, &b);
     MergeVectors(TS, CB, P, NN, a, b);
     RecalculateDistances(CB, NN);
     }
  deallocate(NN);
}

